.model small, c
.386
.stack  

WAIT_FOR_ESC MACRO
    xor   ax, ax
    .WHILE al != 27
      mov   ah, 0
      int   16h
    .ENDW  
ENDM

NEWLINE MACRO
    mov   bx, 0
    mov   ah, 0eh
    mov   al, 13
    int   10h
    mov   bx, 0
    mov   ah, 0eh
    mov   al, 10
    int   10h
ENDM    

read_cmos   PROTO :BYTE
write_cmos  PROTO :BYTE,:BYTE

.data
.code

    ; Place the data that the handlers need to
    ; access in the code segment.
    prevIRQ0Handler   dd 0
    prevIRQ8Handler   dd 0
    counter0          dw 0
    counter8          dw 0

.startup

    ; Hook interrupt 8 (IRQ0).
    mov   ax, 3508h
    int   21h
    mov   WORD PTR prevIRQ0Handler, bx
    mov   WORD PTR prevIRQ0Handler[2], es
    push  ds
    mov   ax, 2508h
    push  cs
    pop   ds
    mov   dx, OFFSET IRQ0Handler
    int   21h
    pop   ds

    ; Hook interrupt 70h (IRQ8).
    mov   ax,3570h
    int   21h
    mov   WORD PTR prevIRQ8Handler, bx
    mov   WORD PTR prevIRQ8Handler[2], es
    push  ds
    mov   ax, 2570h
    push  cs
    pop   ds
    mov   dx,OFFSET IRQ8Handler
    int   21h
    pop   ds
    
    ; Unmask IRQ8.
    in    al, 0A1h
    and   al, not 1
    out   0A1h, al

    WAIT_FOR_ESC

    ; Enable the update complete interrupt.
    invoke read_cmos, 0Bh     ; Get value of status register B.
    push  ax                  ; Preserve it.
    or    al, 10h             ; Set bit 4.
    invoke write_cmos, 0Bh, al    

    NEWLINE
    WAIT_FOR_ESC

    ; Disable the update complete interrupt.
    pop   ax                  ; Recover original value.
    invoke write_cmos,0Bh,al  ; Write it back.

    NEWLINE
    WAIT_FOR_ESC
    NEWLINE

    ; Enable the periodic interrupt.
    invoke read_cmos, 0Bh     ; Get value of status register B.
    push  ax                  ; Preserve it.
    or    al, 40h             ; Set bit 6.
    invoke write_cmos, 0Bh, al

    WAIT_FOR_ESC    

    ; Disable the periodic interrupt.    
    pop   ax                  ; Recover original value.
    invoke write_cmos,0Bh,al  ; Write it back.

    NEWLINE
    WAIT_FOR_ESC
    NEWLINE    

    ; Load the hours, minutes, and seconds alarm registers
    ; with the "wildcard" don't-care value (so the registers
    ; will be ignored when determining the alarm time).
    ; This will cause the RTC to generate an alarm interrupt
    ; for each second.
    invoke write_cmos,5,0ffh    ; Alarm hour.
    invoke write_cmos,3,0ffh    ; Alarm minute.
    invoke write_cmos,1,0ffh    ; Alarm second.
    
    ; Enable the alarm interrupt.
    invoke read_cmos, 0Bh     ; Get value of status register B.
    push  ax                  ; Preserve it.
    or    al, 20h             ; Set bit 5.
    invoke write_cmos, 0Bh, al

    WAIT_FOR_ESC

    ; Disable the alarm interrupt.
    pop   ax                  ; Recover original value.
    invoke write_cmos,0Bh,al  ; Write it back.
    
    NEWLINE
    WAIT_FOR_ESC

    ; Unhook and exit.
    push  ds
    lds   dx, prevIRQ0Handler
    mov   ax, 2508h
    int   21h
    pop   ds

    push  ds
    lds   dx, prevIRQ8Handler
    mov   ax, 2570h
    int   21h
    pop   ds

.exit

; This will display once per 18 ticks (~1 second), to provide
; a rough indicator of the interrupt rates for Interrupt 70h.
IRQ0Handler proc

    sti    
    inc   cs:counter0
    .IF (cs:counter0 > 18)
      mov   cs:counter0, 0
      push  ax
      push  bx
      ; Display a 'T'.
      mov   bx, 0
      mov   ah, 0eh
      mov   al, 'T'
      int   10h
      pop   bx
      pop   ax
    .ENDIF
    ; Chain to the previous handler.
    jmp   cs:prevIRQ0Handler 

IRQ0Handler endp

IRQ8Handler proc

    sti
    push  ax
    push  bx

    ; Read status register C to determine what event caused
    ; the interrupt. This read is necessary for the interrupt
    ; to be called again.
    invoke read_cmos, 0Ch
    .IF al & 10h
      ; Display a 'U'.
      mov   bx, 0
      mov   ah, 0eh
      mov   al, 'U'
      int   10h
    .ELSEIF al & 20h
      ; Display a 'A'.
      mov   bx, 0
      mov   ah, 0eh
      mov   al, 'A'
      int   10h
    .ELSEIF al & 40h
      inc   cs:counter8
      ; This will display once every 1024 interrupts.
      .IF (cs:counter8 > 1024)
        mov   cs:counter8, 0
        ; Display a 'P'.
        mov   bx, 0
        mov   ah, 0eh
        mov   al, 'P'
        int   10h
      .ENDIF
    .ELSE
        ; Display a '?'.
        mov   bx, 0
        mov   ah, 0eh
        mov   al, '?'
        int   10h
    .ENDIF

    ; Not chaining to the previous handler produced
    ; no apparent changes in behavior.    
    ;mov   al, 20h       ; Send an EOI command
    ;out   0A0h, al      ;  to the slave PIC,
    ;out   020h, al      ;  and the master.
    ;pop   bx
    ;pop   ax
    ;iret

    pop   bx
    pop   ax

    ; Chain to the previous handler.
    jmp   cs:prevIRQ8Handler

IRQ8Handler endp
    
; 
;
; This proc returns the value from the specified CMOS address
; in AL.
;
; Bit 7 of the CMOS address port is used to control the
; state of the NMI (Nonmaskable Interrupt). The remaining
; 7 bits are used to specify one of 128 possible addresses.
;
; The state of the NMI disable cannot be determined, so this
; proc assumes that the NMI was enabled, and re-enables it
; before exit.
;
; For the time and date registers this procedure implements
; the logic necessary to reliably read these registers.
;
; In practice, for anything resembling a recent system, the
; I/O delay is probably not necessary because the chipset
; inserts its own delay.
;
; This proc is based mostly on the read_CMOS example, and
; related information, from Frank van Gilluwe's book The
; Undocumented PC, First Edition, Addison-Wesley, 1994,
; ISBN 0-201-62277-7.
;
; The detail of setting the address port to status register
; D and doing a final dummy read of the data port before was
; recommended here:
;
;  http://www.compuphase.com/int70.txt
;
; 

IODELAY MACRO
    REPEAT 10
      jmp   $+2
    ENDM
ENDM

read_cmos proc c address:BYTE

    cli                       ; Disable hardware interrupts.

    .IF address < 0Ah
        mov   al, 0Ah         ; Address of status register A.
        or    al, 80h         ; Set bit 7 to disable NMI.
        out   70h, al         ; Write to address port.
      @@:
        IODELAY
        in    al, 71h         ; Wait until update-in-progress
        test  al, 80h         ;  bit (bit 7) is set.
        jz    @B
        mov   al, 0Ah         ; This second write is necessary
        out   70h, al         ;  for some systems.
      @@:
        IODELAY
        in    al, 71h         ; Wait until update-in-progress
        test  al, 80h         ;  bit is clear.
        jnz   @B
    .ENDIF
    
    mov   al, address
    or    al, 80h             ; Set bit 7 to disable NMI.
    out   70h, al             ; Write to address port.
    IODELAY
    in    al, 71h             ; Read value from data port.
    IODELAY
    mov   ah, al              ; Preserve value.
    mov   al, 0Dh             ; Set the address port to status
    out   70h, al             ;  register D and enable NMI.
    in    al, 71h             ; Do dummy read.
    shr   ax, 8               ; Set return value.
    sti                       ; Enable hardware interrupts.
    ret

read_cmos endp

; 
; 
; CAUTION: THIS PROCEDURE, IF IT IS IMPROPERLY USED, OR IF
; IT CONTAINS ERRORS, CAN RENDER YOUR SYSTEM INOPERABLE.
; USE AT YOUR OWN RISK!
;
; This proc writes the specified value to the specified CMOS
; address.
;
; See the NMI-related comments for the read_cmos proc.
;
; See the I/O delay comments for the read_cmos proc.
;
; For the time and date registers this procedure temporarily
; halts the clock to prevent interference from a clock update.
;
; This proc is based mostly on the write_CMOS example, and
; related information, from Frank van Gilluwe's book The
; Undocumented PC, First Edition, Addison-Wesley, 1994,
; ISBN 0-201-62277-7.
;
; The author advises against updating the time or date within
; 2 seconds of a month rollover, or within 2 seconds of a
; DLS change.
; 
; 

write_cmos proc c uses ax dx address:BYTE, value:BYTE    

    cli                       ; Disable hardware interrupts.

    .IF address < 0Ah
        mov   al, 0Bh         ; Address of status register B.
        or    al, 80h         ; Set bit 7 to disable NMI.
        out   70h, al         ; Write to address port.
        IODELAY
        in    al, 71h         ; Get current value.
        mov   dl, al          ; Preserve it.
        IODELAY
        or    al, 80h         ; Set bit 7 to halt clock.
        out   71h, al         ; Write it back.
    .ENDIF

    mov   al, address
    or    al, 80h             ; Set bit 7 to disable NMI.
    out   70h, al             ; Write to address port.
    IODELAY
    mov   al, value
    out   71h, al             ; Write value to data port.
    IODELAY

    .IF address < 0Ah
        mov   al, 0Bh         ; Address of status register B.
        out   70h, al         ; Write to address port.
        IODELAY
        mov   al, dl          ; Recover original value.
        out   71h, al         ; Write it back.
        IODELAY
    .ENDIF

    mov   al, 0Dh             ; Set the address port to status
    out   70h, al             ;  register D and enable NMI.
    in    al, 71h             ; Do dummy read.
    sti                       ; Enable hardware interrupts.
    ret

write_cmos endp

; 
end
